"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initGenerator = initGenerator;
const devkit_1 = require("@nx/devkit");
const tsquery_1 = require("@phenomnomnominal/tsquery");
const enquirer_1 = require("enquirer");
const node_path_1 = require("node:path");
const utils_1 = require("../utils");
const version_1 = require("../version");
async function initGenerator(tree, { skipGenerateExperience = false } = {}) {
    devkit_1.logger.log('Initializing Angular Three...');
    const packageJson = (0, devkit_1.readJson)(tree, 'package.json');
    const version = packageJson['dependencies']?.['angular-three'] ||
        packageJson['devDependencies']?.['angular-three'] ||
        version_1.ANGULAR_THREE_VERSION;
    (0, devkit_1.addDependenciesToPackageJson)(tree, { 'angular-three': version, three: version_1.THREE_VERSION, ngxtension: version_1.NGXTENSION_VERSION }, { '@types/three': version_1.THREE_TYPE_VERSION });
    devkit_1.logger.info('Turning on skipLibCheck...');
    const tsConfigPath = tree.exists('tsconfig.base.json') ? 'tsconfig.base.json' : 'tsconfig.json';
    (0, devkit_1.updateJson)(tree, tsConfigPath, (json) => {
        if (!('skipLibCheck' in json.compilerOptions) || json.compilerOptions?.skipLibCheck === false) {
            json.compilerOptions.skipLibCheck = true;
        }
        return json;
    });
    (0, utils_1.addMetadataJson)(tree, 'angular-three/metadata.json');
    if (skipGenerateExperience) {
        await (0, devkit_1.formatFiles)(tree);
        return () => {
            (0, devkit_1.installPackagesTask)(tree);
        };
    }
    const { generateExperience } = await (0, enquirer_1.prompt)({
        type: 'select',
        name: 'generateExperience',
        message: 'Generate an Experience component?',
        choices: [
            { value: 'append', name: 'append', message: 'Append <ngt-canvas /> to AppComponent template' },
            { value: 'replace', name: 'replace', message: 'Replace AppComponent template with <ngt-canvas />' },
            { value: 'none', name: 'none', message: 'Do not generate an Experience component' },
        ],
        initial: 2,
    });
    if (generateExperience !== 'none') {
        const projects = (0, devkit_1.getProjects)(tree);
        const applicationProjects = Array.from(projects.entries()).reduce((acc, [projectName, project]) => {
            if (project.projectType === 'application') {
                acc.push({ name: projectName, ...project });
            }
            return acc;
        }, []);
        let selectedProject;
        if (applicationProjects.length === 1) {
            selectedProject = applicationProjects[0].name;
        }
        else {
            // prompt
            const { projectName } = await (0, enquirer_1.prompt)({
                type: 'select',
                name: 'projectName',
                message: 'Add Experience to which project?',
                choices: applicationProjects.map((appProject) => ({ value: appProject.name, name: appProject.name })),
            });
            selectedProject = projectName;
        }
        if (!selectedProject) {
            return warnExperienceWasNotGenerated(tree, 'Unable to locate a project');
        }
        const projectConfiguration = (0, devkit_1.readProjectConfiguration)(tree, selectedProject);
        let sourceRoot = projectConfiguration.sourceRoot;
        if (projectConfiguration['prefix']) {
            sourceRoot += '/' + projectConfiguration['prefix'];
        }
        if (!sourceRoot) {
            return warnExperienceWasNotGenerated(tree, `sourceRoot for ${selectedProject} was not found`);
        }
        // generate Experience component
        (0, devkit_1.generateFiles)(tree, (0, node_path_1.join)(__dirname, 'files'), sourceRoot, { tmpl: '' });
        const appComponentPath = `${sourceRoot}/app.component.ts`;
        const exist = tree.exists(appComponentPath);
        const appComponentTemplatePath = `${sourceRoot}/app.component.html`;
        const templateExist = tree.exists(appComponentTemplatePath);
        const appComponentContent = exist ? tree.read(appComponentPath, 'utf8') : null;
        const appComponentTemplateContent = templateExist ? tree.read(appComponentTemplatePath, 'utf8') : null;
        if (!appComponentContent) {
            return warnExperienceWasNotGenerated(tree, `AppComponent was not found`);
        }
        // standalone is true or not exist
        const isStandalone = appComponentContent.includes(`standalone: true`) || !appComponentContent.includes('standalone');
        if (!isStandalone) {
            return warnExperienceWasNotGenerated(tree, `AppComponent is not a Standalone Component`);
        }
        let updatedContent = tsquery_1.tsquery.replace(appComponentContent, 'PropertyAssignment:has(Identifier[name="imports"]) ArrayLiteralExpression', (node) => {
            return `[${node.elements.map((element) => element['escapedText']).join(', ')}, NgtCanvas]`;
        });
        updatedContent = tsquery_1.tsquery.replace(updatedContent, 'SourceFile', (node) => {
            return `
import { NgtCanvas } from 'angular-three';
import { Experience } from './experience/experience.component';

${node.getFullText()}
`;
        });
        const appComponentContentAst = tsquery_1.tsquery.ast(appComponentContent);
        const classMembersQuery = 'ClassDeclaration > :matches(PropertyDeclaration,MethodDeclaration)';
        const members = tsquery_1.tsquery.match(appComponentContentAst, classMembersQuery);
        if (members.length === 0 || generateExperience === 'replace') {
            updatedContent = tsquery_1.tsquery.replace(updatedContent, 'ClassDeclaration', (node) => {
                const withoutBraces = node.getFullText().slice(0, -1);
                return `
${withoutBraces}
  sceneGraph = Experience;
}`;
            });
        }
        else {
            updatedContent = tsquery_1.tsquery.replace(updatedContent, classMembersQuery, (node) => {
                return `
sceneGraph = Experience;
${node.getFullText()}`;
            });
        }
        if (appComponentTemplateContent) {
            const updatedTemplateContent = generateExperience === 'append'
                ? `
${appComponentTemplateContent}
<ngt-canvas [sceneGraph]="sceneGraph" />`
                : `<ngt-canvas [sceneGraph]="sceneGraph" />`;
            tree.write(appComponentTemplatePath, updatedTemplateContent);
        }
        else {
            updatedContent = tsquery_1.tsquery.replace(updatedContent, 'PropertyAssignment:has(Identifier[name="template"]) NoSubstitutionTemplateLiteral', (node) => {
                const fullText = node.getFullText().trim().slice(1, -1);
                return generateExperience === 'append'
                    ? `\`
${fullText}
<ngt-canvas [sceneGraph]="sceneGraph" />\``
                    : `\`<ngt-canvas [sceneGraph]="sceneGraph" />\``;
            });
        }
        tree.write(appComponentPath, updatedContent);
    }
    await (0, devkit_1.formatFiles)(tree);
    return () => {
        (0, devkit_1.installPackagesTask)(tree);
    };
}
function warnExperienceWasNotGenerated(tree, message) {
    devkit_1.logger.warn(`[NGT] ${message}. Angular Three was initialized successfully, but an Experience component was not generated.`);
    return () => {
        (0, devkit_1.installPackagesTask)(tree);
    };
}
exports.default = initGenerator;
//# sourceMappingURL=generator.js.map